/* FILE: avf2ppm.cc                 -*-Mode: c++-*-
 *
 * Program to try to convert from Any (known) Vector Field file
 * to a ppm bitmap file.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "oc.h"
#include "nb.h"
#include "vf.h"

#include "bitmap.h"
#include "colormap.h"
#include "display.h"

/* End includes */

#define AVF2PPM_MAJOR_VERSION	1
#define AVF2PPM_MINOR_VERSION	1
#define AVF2PPM_RELEASE_LEVEL	0
#define AVF2PPM_RELEASE_SERIAL	1

#define AVF2PPM_VERSION OC_MAKE_VERSION(AVF2PPM)

///////////////////////////////////////////////////////////////////////////
struct MiscConfiguration
{
  Nb_DString background_color;
  BOOL drawboundary;
  int margin;
  int rotation;  // Rotation, in degrees.  Should be 0, 90, 180 or 270.
  int width,height;
  REAL4 zoom;
  BOOL crop;
};

int MakeUniqueName(Tcl_Interp *interp,const char* filename,
		   Nb_DString &uniquename)
{ // Uses the Oc_UniqueFilename proc to generate a name
  // that does not exist on the filesystem.  On success, 1
  // is returned and the result is copied into uniquename.
  // Otherwise, 0 is returned.
  Oc_AutoBuf ab1,ab2;
  char *argv[2];
  argv[0]=ab1("Oc_UniqueFilename");  argv[1]=ab2(filename);
  char *cmd=Tcl_Merge(2,argv);
  int errcode=Tcl_Eval(interp,cmd);
  Tcl_Free(cmd);
  if(errcode!=TCL_OK || interp->result[0]=='\0') return 0;
  uniquename=interp->result;
  return 1;
}

int Tcl_AppInit(Tcl_Interp *interp)
{
  Vf_Mesh *mesh;
  Vf_EmptyMesh empty_mesh; // Dummy
  DisplayFrame frame;
  OommfBitmap bitmap;
  PlotConfiguration arrow_config,pixel_config;
  MiscConfiguration misc_config;
  Oc_AutoBuf ab1,ab2,ab3;

  Oc_SetDefaultTkFlag(1);
  if (Oc_Init(interp) == TCL_ERROR) {
    fprintf(stderr,"%s", Tcl_GetStringResult(interp));
    return Tcl_GlobalEval(interp, ab1("exit 1"));
  }
  if (Vf_Init(interp) == TCL_ERROR) {
    fprintf(stderr,"%s", Tcl_GetStringResult(interp));
    return Tcl_GlobalEval(interp, ab1("exit 1"));
  }

  Tcl_PkgProvide(interp, ab1("Avf2ppm"), ab2(AVF2PPM_VERSION));

  if (Oc_InitScript(interp, "Avf2ppm", AVF2PPM_VERSION) != TCL_OK) {
    fprintf(stderr,"InitScript error: %s\n", Tcl_GetStringResult(interp));
    return Tcl_GlobalEval(interp, ab1("exit 1"));
  }

  // Get configuration values from Tcl array
  int colorcount;
  Nb_DString colormapname;
  arrow_config.displaystate=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("arrow,status"),TCL_GLOBAL_ONLY));
  arrow_config.colorquantity=Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("arrow,quantity"),TCL_GLOBAL_ONLY);
  colormapname=Tcl_GetVar2(interp,ab1("plot_config"),
			   ab2("arrow,colormap"),TCL_GLOBAL_ONLY);
  colorcount=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			      ab2("arrow,colorcount"),TCL_GLOBAL_ONLY));
  arrow_config.colormap.Setup(colorcount,colormapname);
  arrow_config.autosample=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("arrow,autosample"),TCL_GLOBAL_ONLY));
  arrow_config.subsample=Nb_Atof(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("arrow,subsample"),TCL_GLOBAL_ONLY));
  arrow_config.mag=Nb_Atof(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("arrow,mag"),TCL_GLOBAL_ONLY));
  arrow_config.antialias=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("arrow,antialias"),TCL_GLOBAL_ONLY));

  pixel_config.colorquantity=Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("pixel,quantity"),TCL_GLOBAL_ONLY);
  colormapname=Tcl_GetVar2(interp,ab1("plot_config"),
			   ab2("pixel,colormap"),TCL_GLOBAL_ONLY);
  colorcount=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			      ab2("pixel,colorcount"),TCL_GLOBAL_ONLY));
  pixel_config.colormap.Setup(colorcount,colormapname);
  pixel_config.displaystate=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("pixel,status"),TCL_GLOBAL_ONLY));
  pixel_config.autosample=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("pixel,autosample"),TCL_GLOBAL_ONLY));
  pixel_config.subsample=Nb_Atof(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("pixel,subsample"),TCL_GLOBAL_ONLY));
  pixel_config.mag=Nb_Atof(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("pixel,mag"),TCL_GLOBAL_ONLY));

  misc_config.background_color=Tcl_GetVar2(interp,ab1("plot_config"),
  			     ab2("misc,background"),TCL_GLOBAL_ONLY);
  misc_config.drawboundary=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("misc,drawboundary"),TCL_GLOBAL_ONLY));
  misc_config.margin=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("misc,margin"),TCL_GLOBAL_ONLY));
  misc_config.width=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("misc,width"),TCL_GLOBAL_ONLY));
  misc_config.height=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("misc,height"),TCL_GLOBAL_ONLY));
  misc_config.zoom=atof(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("misc,zoom"),TCL_GLOBAL_ONLY));
  misc_config.crop=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("misc,crop"),TCL_GLOBAL_ONLY));
  misc_config.rotation=atoi(Tcl_GetVar2(interp,ab1("plot_config"),
			     ab2("misc,rotation"),TCL_GLOBAL_ONLY));

  // Get options from a2pOpts array
  int verbose=atoi(Tcl_GetVar2(interp,ab1("a2pOpts"),
			       ab2("verbose"),TCL_GLOBAL_ONLY));
  int force=atoi(Tcl_GetVar2(interp,ab1("a2pOpts"),
			     ab2("force"),TCL_GLOBAL_ONLY));
  Nb_DString ppm_type(Tcl_GetVar2(interp,ab1("a2pOpts"),
				ab2("type"),TCL_GLOBAL_ONLY));
  Nb_DString filtercmd(Tcl_GetVar2(interp,ab1("a2pOpts"),
				   ab2("filter"),TCL_GLOBAL_ONLY));


  // Get input and output file lists
  char *ilist=Tcl_GetVar2(interp,ab1("a2pOpts"),
			   ab2("ifiles"),TCL_GLOBAL_ONLY);
  char *olist=Tcl_GetVar2(interp,ab1("a2pOpts"),
			  ab2("ofiles"),TCL_GLOBAL_ONLY);
  int icount,ocount;
  char **ifiles,**ofiles;
  Tcl_SplitList(interp,ilist,&icount,&ifiles);
  Tcl_SplitList(interp,olist,&ocount,&ofiles);
  if(ocount<1) {
    // At least one output filename is required
    fprintf(stderr,"Empty output list\n");
    return Tcl_GlobalEval(interp, ab1("exit 2"));
  }

  // If in very verbose mode, dump configuration info to stderr
  if(verbose>1) {
    fprintf(stderr,
	    "Output file type: %s\n",
	    ppm_type.GetStr());
    if(filtercmd.Length()>0) {
      fprintf(stderr,
	      "Filter command: %s\n",
	      filtercmd.GetStr());
    }
    fprintf(stderr,
	    "Arrow Configuration---\n"
	    " Display: %d\n"
	    " Colormap name: %s\n"
	    " Colormap size: %d\n"
	    " Quantity: %s\n"
	    " Autosample: %d\n"
	    " Subsample: %g\n"
	    " Mag: %g\n"
	    " Antialias: %d\n",
	    arrow_config.displaystate,
	    arrow_config.colormap.GetMapName(),
	    arrow_config.colormap.GetColorCount(),
	    arrow_config.colorquantity.GetStr(),
	    arrow_config.autosample,
	    arrow_config.subsample,
	    arrow_config.mag,
	    arrow_config.antialias);
    fprintf(stderr,
	    "Pixel Configuration---\n"
	    " Display: %d\n"
	    " Colormap name: %s\n"
	    " Colormap size: %d\n"
	    " Quantity: %s\n"
	    " Autosample: %d\n"
	    " Subsample: %g\n"
	    " Mag: %g\n",
	    pixel_config.displaystate,
	    pixel_config.colormap.GetMapName(),
	    pixel_config.colormap.GetColorCount(),
	    pixel_config.colorquantity.GetStr(),
	    pixel_config.autosample,
	    pixel_config.subsample,
	    pixel_config.mag);
    fprintf(stderr,
	    "Misc Configuration----\n"
	    " Background color: %s\n"
	    " Draw boundary: %d\n"
	    " Margin: %d\n"
	    " Dimensions: %d x %d\n"
	    " Zoom: %g\n"
	    " Crop: %d\n"
	    " Rotation: %d\n",
	    misc_config.background_color.GetStr(),
	    misc_config.drawboundary,
	    misc_config.margin,
	    misc_config.width,
	    misc_config.height,
	    misc_config.zoom,
	    misc_config.crop,
	    misc_config.rotation);
  }

  // Base frame configuration
  frame.SetTclInterp(interp);
  frame.SetPlotConfiguration(arrow_config,pixel_config);
  frame.SetBackgroundColor((const char *)misc_config.background_color);
  frame.SetDrawBoundary(misc_config.drawboundary);
  int rotation_index=int(OC_ROUND(misc_config.rotation/90.))%4;
  if(rotation_index<0) rotation_index+=4;
  switch(rotation_index)
    {
    case 1:  frame.SetCoordinates(CS_DISPLAY_ROT_90);   break;
    case 2:  frame.SetCoordinates(CS_DISPLAY_ROT_180);  break;
    case 3:  frame.SetCoordinates(CS_DISPLAY_ROT_270);  break;
    default: frame.SetCoordinates(CS_DISPLAY_STANDARD); break;
    }
  misc_config.rotation=rotation_index*90;

  // Loop through input files
  Nb_DString outstr;
  for(int i=0;i<icount;i++) {
    const char *inname=ifiles[i];
    const char *outname;
    if(i<ocount) outname=ofiles[i];
    else         outname=ofiles[ocount-1];
    if(!force) {
      if(MakeUniqueName(interp,outname,outstr)==0) {
	fprintf(stderr,"ERROR: Unable to generate unique name based on %s\n",
		outname);
	continue;
      }
      outname=outstr.GetStr();
    }

    // Fill mesh from input file
    Vf_FileInput *vffreader = Vf_FileInput::NewReader(inname);
    if(vffreader==NULL) { // Unknown file type or other error
      fprintf(stderr,"ERROR detected processing file %s\n",inname);
      continue;
    }
    else { // Reader found for inname
      mesh=vffreader->NewMesh();
      delete vffreader;
      if(mesh==NULL) {
	fprintf(stderr,"ERROR detected while reading file %s\n",inname);
	continue;
      }
    }
    
    // Initialize DisplayFrame
    frame.SetMesh(mesh);

    // Configure bitmap
    if(misc_config.zoom==0) {
      frame.SetZoom(misc_config.width-2*misc_config.margin,
		    misc_config.height-2*misc_config.margin);
    } else {
      frame.SetZoom(misc_config.zoom);
    }
    Nb_BoundingBox<REAL4> bbox=frame.GetDisplayBox();
    Nb_Vec3<REAL4> minpt,maxpt;
    bbox.GetExtremes(minpt,maxpt);
    int xmin,ymin,xmax,ymax,xmarginadj,ymarginadj;
    xmin=int(floor(minpt.x))-misc_config.margin;
    ymin=int(floor(minpt.y))-misc_config.margin;
    xmax=int(ceil(maxpt.x))+misc_config.margin;
    ymax=int(ceil(maxpt.y))+misc_config.margin;
    xmarginadj=(xmax-xmin+1)-misc_config.width;
    ymarginadj=(ymax-ymin+1)-misc_config.height;
    if(xmarginadj>0) {
      // Resulting size is bigger than requested. Trim.
      xmin+=int(floor(xmarginadj/2.));
      xmax-=int(ceil(xmarginadj/2.));
    }
    if(ymarginadj>0) {
      // Resulting size is bigger than requested. Trim.
      ymin+=int(floor(ymarginadj/2.));
      ymax-=int(ceil(ymarginadj/2.));
    }
    if(!misc_config.crop) {
      // No-crop request, so make bitmap exact size requested
      xmax=xmin+misc_config.width-1;
      ymax=ymin+misc_config.height-1;
    }
    bitmap.Setup(xmin,ymin,xmax,ymax);    // Ignore z

    // Dump file processing information to stderr, if requested
    if(verbose>0) {
      if(filtercmd.Length()==0) {
	fprintf(stderr," %s --> %s",inname,outname);
      } else {
	fprintf(stderr," %s --> %s --> %s",inname,filtercmd.GetStr(),outname);
      }
      bitmap.GetExtents(xmin,ymin,xmax,ymax);
      fprintf(stderr,"  (%dx%d)\n",xmax-xmin+1,ymax-ymin+1);
    }

    // Fill bitmap
    frame.FillBitmap(bitmap);
    
    // Write bitmap
    Tcl_Channel channel;
    if(filtercmd.Length()==0) {
      // Dump output direct to a file
      channel=Tcl_OpenFileChannel(interp,ab1(outname),ab2("w"),0666);
      if(channel==NULL) {
	fprintf(stderr,"ERROR: Unable top open output file %s\n",outname);
	fprintf(stderr,"   %s\n",interp->result);
      }
    } else {
      // Open command channel for filtering
      channel=NULL;
      Nb_DString cmd(filtercmd);
      cmd.Append(" > {");
      cmd.Append(outname);
      cmd.Append("}");      // Protect spaces in outname
      int cmdc; char** cmdv;
      if(Tcl_SplitList(interp,ab1(cmd.GetStr()),&cmdc,&cmdv)==TCL_OK) {
	channel=Tcl_OpenCommandChannel(interp,cmdc,cmdv,TCL_STDIN);
	Tcl_Free((char *)cmdv);
      }
      if(channel==NULL) {
	fprintf(stderr,"   %s\n",interp->result);
      }
    }
    if(channel!=NULL) {
      Tcl_SetChannelOption(interp,channel,
                ab1("-buffersize"),ab2("10000"));
      Tcl_SetChannelOption(interp,channel,
                ab1("-buffering"),ab2("full"));
      Tcl_SetChannelOption(interp,channel,
                ab1("-translation"),ab2("binary"));
      if(strcmp("P6",ppm_type.GetStr())==0)
	bitmap.WritePPMChannel(channel,6);
      else 
	bitmap.WritePPMChannel(channel,3);
      Tcl_Close(interp,channel);
    }

    // Clean up
    frame.SetMesh(&empty_mesh);
    delete mesh;
  }

  // That's all folks!
  Tcl_Free((char *)ifiles);
  Tcl_Free((char *)ofiles);

  return Tcl_GlobalEval(interp, ab1("exit 0"));
}

int Oc_AppMain(int argc,char **argv)
{
#define DUMMY_SCRIPTNAME "foo"
  // Add a dummy argv[1] for Oc_Main to eat as fake script name.
  int my_argc=argc+1;
  char** my_argv=new char*[my_argc];
  my_argv[1]=new char[strlen(DUMMY_SCRIPTNAME)+1];
  strcpy(my_argv[1],DUMMY_SCRIPTNAME);
  my_argv[0]=argv[0];
  for(int i=2;i<my_argc;i++) my_argv[i]=argv[i-1];

  Oc_Main(my_argc, my_argv, Tcl_AppInit);

  // Oc_Main doesn't return, but for reference here is
  // the memory release code.
  delete[] my_argv[0];
  delete[] my_argv;
  return 0;
#undef DUMMY_SCRIPTNAME
}
