/* FILE: bitmap.h                 -*-Mode: c++-*-
 *
 * OOMMF C++ classes for bitmap rendering
 *
 */

#ifndef OBITMAP_H
#define OBITMAP_H

#include "oc.h"
#include "nb.h"
#include "vf.h"

#include "colormap.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// OOMMF Polygon class
// This class defines a closed (not necessarily convex) polygon, lying
// in the x-y plane.  It is used by the bitmap drawing routines.
struct PlanePoint {
public:
  REAL8m x,y;
  PlanePoint() { x=y=0.; }
  PlanePoint(REAL8m xval,REAL8m yval) { x=xval; y=yval; }
  REAL8m Mag(void) { return hypot(x,y); }
  REAL8m MagSq(void) { return x*x+y*y; }
};

class OommfPolygon {
private:
  static const ClassDoc class_doc;
  Nb_List<PlanePoint> vertex;
public:
  OommfPolygon() {}
  OommfPolygon(const OommfPolygon& poly); // Implicit copy constructor
  ~OommfPolygon() {}
  OommfPolygon& operator=(OommfPolygon& poly);
  void Clear(); // Removes all vertices from polygon (does not free memory)
  int GetVertexCount();
  void AddVertex(const PlanePoint& pt); // Adds a new vertex to polygon
  void GetBounds(REAL8m& xmin,REAL8m& ymin,
                 REAL8m& xmax,REAL8m& ymax) const;
  int IsInside(const PlanePoint& pt) const;
  /// Returns 1 if pt is inside the polygon
  void Scale(REAL8m xscale,REAL8m yscale);    // Coordinate
  void Shift(REAL8m xoffset,REAL8m yoffset);  //    transformations
  void Rotate(REAL8m angcos,REAL8m angsin);   //  "  "
  void AdjustFirstToInteger(); // Rotates and rescales bitmap as
  /// necessary to shift first vertex point to the nearest integral
  /// coordinate
  void AdjustAllToIntegers(); // Rounds position of each vertex
  /// individually to nearest integer.

  void DumpVertices(FILE* fout);  // Debugging tool
};

//////////////////////////////////////////////////////////////////////////
// OOMMF Bitmap class
class OommfBitmap {
private:
  static const ClassDoc class_doc;
  int xmin,ymin,xmax,ymax;  // Inclusive
  int xsize,ysize,totalsize; // xsize=xmax-xmin+1, etc.
  OommfPackedRGB* bitmap; // Rectangular display region
  void Alloc(int new_xmin,int new_ymin,
	     int new_xmax,int new_ymax);
  OommfPolygon default_arrow,default_diamond,default_cross;
  OommfPolygon work_poly; // Class scratch space.  Used by
  /// the following routines: DrawFilledArrow, DrawDiamondWithCross.
  /// Must not be used by any routine calling either of these.
  void Free();
  OommfPackedRGB GetPixel(int x,int y) const;
  void SetPixel(int x,int y,OommfPackedRGB color);
  /// Both GetPixel and SetPixel ASSUME x,y is in-bounds
public:
  OommfBitmap();
  void Setup(int new_xmin,int new_ymin,int new_xmax,int new_ymax);
  ~OommfBitmap() { Free(); }

  void GetExtents(int& myxmin,int& myymin,int& myxmax,int& myymax) const {
    myxmin=xmin; myymin=ymin; myxmax=xmax; myymax=ymax;
  }

  int WritePPMFile(const char* filename,int ppm_type);
  /// Use filename == NULL or "-" to write to stdout.
  /// "type" corresponds to the PPM magic number.  It should be either
  ///   3 (ASCII text) or 6 (binary).
  /// Return value is
  ///     0 => No error
  ///     1 => Bad parameter
  ///     2 => Unable to open filename for output
  ///     4 => Write error

  int WritePPMChannel(Tcl_Channel chan,int ppm_type);
  /// Analogous to WritePPMFile, but writes to a previously
  /// opened Tcl_Channel.  Returns 0 on success, >0 on error.

  // Drawing routines
  void Erase(OommfPackedRGB color); // Writes 'color' on each pixel

  void DrawFilledRectangle(REAL8m x1,REAL8m y1,REAL8m x2,REAL8m y2,
			  OommfPackedRGB color);
  /// Opposing vertices (x1,y1) and (x2,y2) are included.

  // Draws frame outside specified rectangle.
  void DrawRectangleFrame(REAL8m x1,REAL8m y1,
			  REAL8m x2,REAL8m y2,
			  REAL8m borderwidth,
			  OommfPackedRGB color);

  // Filled, closed polygon.
  void DrawFilledPoly(const OommfPolygon& poly,
		      OommfPackedRGB color,BOOL antialias);

  // Draws (potentially fat) line segments between each successive
  // pair of points in vlist.  If the miter!=0, then the segment
  // ends are mitered.  In general, this is an unclosed figure,
  // unless the last and first point are equal, in which case those
  // ends will be mitered if requested.
  void DrawPolyLine(const Nb_List<PlanePoint>& vlist,
		    REAL8m linewidth,OommfPackedRGB color,
		    int miter,int antialias);

  /////// The next 3 drawing routines are specialized methods for
  /////// rendering vector fields, realized in terms of the
  /////// preceding general drawing routines, DrawFilledRectangle
  /////// and DrawFilledPoly.  It can be argued that these specialized
  /////// routines don't properly belong inside the OommfBitmap class.
  ///////
  void DrawFilledArrow(REAL8m xc,REAL8m yc,REAL8m mag,
                      REAL8m xcos,REAL8m ycos,REAL8m zcos,
                      OommfPackedRGB color,BOOL antialias);
  /// Draws arrow centered at point (xc,yc) in bitmap, of size mag,
  /// with directional cosines xcos,ycos,zcos, in the specified color.

  void DrawDiamondWithCross(REAL8m xc,REAL8m yc,REAL8m mag,
			   OommfPackedRGB outercolor,
			   OommfPackedRGB innercolor);
  /// Symbol representing out-of-plane vectors

  void DrawSquareWithDot(REAL8m xc,REAL8m yc,REAL8m mag,
			 OommfPackedRGB outercolor,
			 OommfPackedRGB innercolor);
  /// Symbol representing in-to-plane vectors.

};

#endif // OBITMAP_H
